##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from dataclasses import dataclass, field
from re import sub
from typing import Any, Dict, List, Optional, Tuple, Type, Callable
from ..argument_types import EmptyArray, Pauli, Range, Result


__all__ = ["InputParams"]


class EntryPointArguments:
    """
    Wrapper class to set QIR entry point arguments.

    This class is used to set QIR entry point arguments inside the
    InputParamsItem class.  It overrides the __setitem__ method to
    automatically detect the entry point's argument from the passed value.
    """

    # Maps supported Python types to QIR entry point type names and a callable
    # that extracts the serialized value from the original value. (list (=
    # Array) is handled as a special case.)
    type_map: Dict[type, Tuple[str, Callable]] = {
        int: ("Int", lambda v: v),
        float: ("Double", lambda v: v),
        bool: ("Boolean", lambda v: v),
        str: ("String", lambda v: v),
        Pauli: ("Pauli", lambda v: v.value),
        Result: ("Result", lambda v: v.value),
        Range: ("Range", lambda v: v.value)
    }

    def __init__(self):
        self.entries = []

    def __setitem__(self, name: str, value: Any):
        """
        Creates entry point argument entry and automatically determines the
        type from value.
        """
        if type(value) == list:
            if len(value) == 0:
                raise ValueError("Use EmptyArray(type) to assign an empty "
                                 "error")
            else:
                first = value[0]
                first_value, first_type = self._extract_value_and_type(first)
                values = [first_value]
                for next in value[1:]:
                    next_value, next_type = self._extract_value_and_type(next)
                    if next_type != first_type:
                        raise TypeError("All elements in a list must have "
                                        "the same type")
                    values.append(next_value)
                self.entries.append(
                    {"name": name,
                     "value": values,
                     "type": "Array",
                     "elementType": first_type})
        elif type(value) == EmptyArray:
            element_type = self._extract_type(value.element_type)
            self.entries.append(
                {"name": name,
                 "value": [],
                 "type": "Array",
                 "elementType": element_type})
        else:
            entry_value, entry_type = self._extract_value_and_type(value)
            self.entries.append(
                {"name": name, "value": entry_value, "type": entry_type})

    def _extract_type(self, type: Type) -> str:
        """
        Convert Python type to QIR entry point argument type name.
        """
        if type in self.type_map:
            return self.type_map[type][0]
        elif type == list:
            raise TypeError(f"Nested lists are not supported")
        else:
            type_name = type.__name__
            raise TypeError(f"Unsupported type {type_name}")

    def _extract_value_and_type(self, value: Any) -> Tuple[Any, str]:
        """
        Convert Python value to QIR entry point argument type name and
        serialized value.
        """
        if type(value) in self.type_map:
            entry_type, entry_value_func = self.type_map[type(value)]
            return entry_value_func(value), entry_type
        elif type(value) == list:
            raise TypeError(f"Nested lists are not supported")
        else:
            type_name = type(value).__name__
            raise TypeError(f"Unsupported type {type_name} for {value}")


@dataclass
class AutoValidatingParams:
    """
    A helper class for target parameters.

    It has a function as_dict that automatically extracts a dictionary from
    the class' fields.  They are added to the result dictionary if their value
    is not None, the key is automatically transformed from Python snake case
    to camel case, and if validate is True and if the field has a validation
    function, the field is validated beforehand.
    """
    def as_dict(self, validate=True):
        result = {}

        for name, field in self.__dataclass_fields__.items():
            field_value = self.__getattribute__(name)
            if field_value is not None:
                # validate field?
                if validate and "validate" in field.metadata:
                    func = field.metadata["validate"]
                    # check for indirect call (like in @staticmethod)
                    if hasattr(func, "__func__"):
                        func = func.__func__
                    func(name, field_value)

                # translate field name to camel case
                s = sub(r"(_|-)+", " ", name).title().replace(" ", "")
                attribute = ''.join([s[0].lower(), s[1:]])
                result[attribute] = field_value

        if validate:
            self.post_validation(result)

        return result

    def post_validation(self, result):
        """
        A function that is called after all individual fields have been
        validated, but before the result is returned.

        Here result is the current dictionary.
        """
        pass


def validating_field(validation_func, default=None):
    """
    A helper method to declare field for an AutoValidatingParams data class.
    """
    return field(default=default, metadata={"validate": validation_func})


class InputParamsItem:
    """
    Base class for input parameters.

    This class serves both as the base class for InputParams as well as for the
    items in the InputParams.
    """
    def __init__(self):
        # all input param items may have an entry point name and a list of
        # arguments
        self.entry_point: Optional[str] = None
        self.arguments = EntryPointArguments()

    def as_dict(self, validate=True) -> Dict[str, Any]:
        """
        Returns input params as a dictionary.
        """
        result = {}

        if self.entry_point is not None:
            result['entryPoint'] = self.entry_point

        if len(self.arguments.entries) > 0:
            result['arguments'] = self.arguments.entries

        return result


class InputParams(InputParamsItem):
    """
    Class to define input parameters.

    This class allows to define input parameters for non-batching and batching
    jobs.  The instance represents a batching job, if and only if num_items is
    set to some positive number less or equal to MAX_NUM_ITEMS.

    Both this class and the items in this class are based on InputParamsItem as
    a template, which can be overriden for specializations created by a target.
    This class should never be constructed directly but only through the
    InputParams.make_params method.
    """

    MAX_NUM_ITEMS: int = 1000

    def __init__(
            self,
            num_items: Optional[int] = None,
            item_type: Type[InputParamsItem] = InputParamsItem):
        """
        Constructs a InputParams instance.

        The item_type argument should be set by targets that override
        InputParams and have a specialized InputParamsItem class.
        """
        item_type.__init__(self)

        # fileURIs
        self.file_uris = {}

        if num_items is not None:
            self.has_items = True
            if num_items <= 0 or num_items > self.MAX_NUM_ITEMS:
                raise ValueError(
                    "num_items must be a positive value less or equal to "
                    f"{self.MAX_NUM_ITEMS}")
            self._items = [item_type() for _ in range(num_items)]
        else:
            self.has_items = False

        self.item_type = item_type

    @property
    def items(self) -> List:
        if self.has_items:
            return self._items
        else:
            raise Exception("Cannot access items in a non-batching job, call "
                            "make_params with num_items parameter")

    def as_dict(self, validate=True) -> Dict[str, Any]:
        """
        Constructs a dictionary from the input params.

        For batching jobs, top-level entries are merged into item entries.
        Item entries have priority in case they are specified.
        """

        # initialize result and set type hint
        result: Dict[str, Any] = self.item_type.as_dict(self, validate)

        if self.has_items:
            result["items"] = [item.as_dict(validate) for item in self._items]
            # In case of batching, no need to stop if failing an item
            result["resumeAfterFailedItem"] = True

        # add fileUris if existing
        if len(self.file_uris) > 0:
            result["fileUris"] = self.file_uris

        return result



import json
from typing import Any, Dict, Union

class JobFailedWithResultsError(RuntimeError):
    """
    Error produced when Job completes with status "Failed" and the Job
    supports producing failure results.

    The failure results can be accessed with get_failure_results() method
    """
    
    def __init__(self, message: str, failure_results: Any, *args: object) -> None:
        self._set_error_details(message, failure_results)
        super().__init__(message, *args)


    def _set_error_details(self, message: str, failure_results: Any) -> None:
        self._message = message
        try:
            decoded_failure_results = failure_results.decode("utf8")
            self._failure_results: Dict[str, Any] = json.loads(decoded_failure_results)
        except:
            self._failure_results = failure_results


    def get_message(self) -> str:
        """
        Get error message.
        """
        return self._message


    def get_failure_results(self) -> Union[Dict[str, Any], str]:
        """
        Get failure results produced by the job.
        """
        return self._failure_results
    

    def __str__(self) -> str:
        return f"{self._message}\nFailure results: {self._failure_results}"


#!/bin/env python
# -*- coding: utf-8 -*-
##
# resource_estimator.py: Qiskit result for microsoft.estimator target.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from azure.quantum.target.microsoft import MicrosoftEstimatorResult


def make_estimator_result(data):
    if not data["success"]:
        error_data = data["error_data"]
        message = "Cannot retrieve results as job execution failed " \
                  f"({error_data['code']}: {error_data['message']})"

        raise RuntimeError(message)
    
    results = data["results"]
    if len(results) == 1:
        data = results[0]['data']
        return MicrosoftEstimatorResult(data)
    else:
        raise ValueError("Expected Qiskit results for RE be of length 1")



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

"""Azure Quantum Cirq Service"""

from .service import AzureQuantumService
from .job import Job

__all__ = ["AzureQuantumService", "Job"]


# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from copy import deepcopy
from typing import Any, TYPE_CHECKING

from azure.core import PipelineClient
from azure.core.rest import HttpRequest, HttpResponse

from . import models as _models
from ._configuration import QuantumClientConfiguration
from ._serialization import Deserializer, Serializer
from .operations import (
    JobsOperations,
    ProvidersOperations,
    QuotasOperations,
    SessionsOperations,
    StorageOperations,
    TopLevelItemsOperations,
)

if TYPE_CHECKING:
    # pylint: disable=unused-import,ungrouped-imports
    from azure.core.credentials import TokenCredential


class QuantumClient:  # pylint: disable=client-accepts-api-version-keyword
    """Azure Quantum REST API client.

    :ivar jobs: JobsOperations operations
    :vartype jobs: azure.quantum._client.operations.JobsOperations
    :ivar providers: ProvidersOperations operations
    :vartype providers: azure.quantum._client.operations.ProvidersOperations
    :ivar storage: StorageOperations operations
    :vartype storage: azure.quantum._client.operations.StorageOperations
    :ivar quotas: QuotasOperations operations
    :vartype quotas: azure.quantum._client.operations.QuotasOperations
    :ivar sessions: SessionsOperations operations
    :vartype sessions: azure.quantum._client.operations.SessionsOperations
    :ivar top_level_items: TopLevelItemsOperations operations
    :vartype top_level_items: azure.quantum._client.operations.TopLevelItemsOperations
    :param subscription_id: The Azure subscription ID. This is a GUID-formatted string (e.g.
     00000000-0000-0000-0000-000000000000). Required.
    :type subscription_id: str
    :param resource_group_name: Name of an Azure resource group. Required.
    :type resource_group_name: str
    :param workspace_name: Name of the workspace. Required.
    :type workspace_name: str
    :param credential: Credential needed for the client to connect to Azure. Required.
    :type credential: ~azure.core.credentials.TokenCredential
    :keyword endpoint: Service URL. Default value is "https://quantum.azure.com".
    :paramtype endpoint: str
    :keyword api_version: Api Version. Default value is "2022-09-12-preview". Note that overriding
     this default value may result in unsupported behavior.
    :paramtype api_version: str
    """

    def __init__(
        self,
        subscription_id: str,
        resource_group_name: str,
        workspace_name: str,
        credential: "TokenCredential",
        *,
        endpoint: str = "https://quantum.azure.com",
        **kwargs: Any
    ) -> None:
        self._config = QuantumClientConfiguration(
            subscription_id=subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            credential=credential,
            **kwargs
        )
        self._client: PipelineClient = PipelineClient(base_url=endpoint, config=self._config, **kwargs)

        client_models = {k: v for k, v in _models._models.__dict__.items() if isinstance(v, type)}
        client_models.update({k: v for k, v in _models.__dict__.items() if isinstance(v, type)})
        self._serialize = Serializer(client_models)
        self._deserialize = Deserializer(client_models)
        self._serialize.client_side_validation = False
        self.jobs = JobsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.providers = ProvidersOperations(self._client, self._config, self._serialize, self._deserialize)
        self.storage = StorageOperations(self._client, self._config, self._serialize, self._deserialize)
        self.quotas = QuotasOperations(self._client, self._config, self._serialize, self._deserialize)
        self.sessions = SessionsOperations(self._client, self._config, self._serialize, self._deserialize)
        self.top_level_items = TopLevelItemsOperations(self._client, self._config, self._serialize, self._deserialize)

    def send_request(self, request: HttpRequest, **kwargs: Any) -> HttpResponse:
        """Runs the network request through the client's chained policies.

        >>> from azure.core.rest import HttpRequest
        >>> request = HttpRequest("GET", "https://www.example.org/")
        <HttpRequest [GET], url: 'https://www.example.org/'>
        >>> response = client.send_request(request)
        <HttpResponse: 200 OK>

        For more information on this code flow, see https://aka.ms/azsdk/dpcodegen/python/send_request

        :param request: The network request you want to make. Required.
        :type request: ~azure.core.rest.HttpRequest
        :keyword bool stream: Whether the response payload will be streamed. Defaults to False.
        :return: The response of your network call. Does not do error handling on your response.
        :rtype: ~azure.core.rest.HttpResponse
        """

        request_copy = deepcopy(request)
        request_copy.url = self._client.format_url(request_copy.url)
        return self._client.send_request(request_copy, **kwargs)

    def close(self) -> None:
        self._client.close()

    def __enter__(self) -> "QuantumClient":
        self._client.__enter__()
        return self

    def __exit__(self, *exc_details: Any) -> None:
        self._client.__exit__(*exc_details)



"""Defines classes for interacting with Microsoft Elements DFT service"""

from .target import MicrosoftElementsDft
from .job import MicrosoftElementsDftJob

__all__ = ["MicrosoftElementsDft", "MicrosoftElementsDftJob"]


# Auto-generated file, do not edit.
##
# version.py: Specifies the version of the azure.quantum package.
##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
__version__ = "1.1.0"



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

from typing import TYPE_CHECKING, Any, Dict, List
from azure.quantum.version import __version__
from qiskit import QuantumCircuit
from abc import abstractmethod
from .backend import AzureQirBackend

from qiskit.providers.models import BackendConfiguration
from qiskit.providers import Options, Provider

QIR_BASIS_GATES = [
    "measure",
    "m",
    "ccx",
    "cx",
    "cz",
    "h",
    "reset",
    "rx",
    "ry",
    "rz",
    "s",
    "sdg",
    "swap",
    "t",
    "tdg",
    "x",
    "y",
    "z",
    "id",
]

if TYPE_CHECKING:
    from azure.quantum.qiskit import AzureQuantumProvider

import logging

logger = logging.getLogger(__name__)

__all__ = ["MicrosoftBackend", "MicrosoftResourceEstimationBackend"]


class MicrosoftBackend(AzureQirBackend):
    """Base class for interfacing with a Microsoft backend in Azure Quantum"""

    @abstractmethod
    def __init__(
        self, configuration: BackendConfiguration, provider: Provider = None, **fields
    ):
        super().__init__(configuration, provider, **fields)

    @classmethod
    def _default_options(cls):
        return Options(targetCapability="AdaptiveExecution")

    def _azure_config(self) -> Dict[str, str]:
        config = super()._azure_config()
        config.update(
            {
                "provider_id": "microsoft-qc",
                "output_data_format": "microsoft.resource-estimates.v1",
                "to_qir_kwargs": {"record_output": False},
            }
        )
        return config


class MicrosoftResourceEstimationBackend(MicrosoftBackend):
    """Backend class for interfacing with the resource estimator target"""

    backend_names = ("microsoft.estimator",)

    @classmethod
    def _default_options(cls):
        return Options(
            targetCapability="AdaptiveExecution",
            errorBudget=1e-3,
            qubitParams={"name": "qubit_gate_ns_e3"},
            qecScheme={"name": "surface_code"}
        )

    def __init__(self, name: str, provider: "AzureQuantumProvider", **kwargs):
        """Constructor for class to interface with the resource estimator target"""
        default_config = BackendConfiguration.from_dict(
            {
                "backend_name": name,
                "backend_version": __version__,
                "simulator": True,
                "local": False,
                "coupling_map": None,
                "description": "Resource estimator on Azure Quantum",
                "basis_gates": QIR_BASIS_GATES,
                "memory": False,
                "n_qubits": 0xFFFFFFFFFFFFFFFF,  # NOTE: maximum 64-bit unsigned value
                "conditional": True,
                "max_shots": 1,
                "max_experiments": 1,
                "open_pulse": False,
                "gates": [
                    {"name": "TODO", "parameters": [], "qasm_def": "TODO"}
                ],  # NOTE: copied from other backends
                "azure": self._azure_config(),
            }
        )
        logger.info("Initializing MicrosoftResourceEstimationBackend")
        configuration: BackendConfiguration = kwargs.pop(
            "configuration", default_config
        )
        super().__init__(configuration=configuration, provider=provider, **kwargs)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from typing import TYPE_CHECKING, Optional, Union, Protocol, List
from abc import abstractmethod

from azure.quantum._client.models import SessionDetails, SessionStatus, SessionJobFailurePolicy
from azure.quantum.job.workspace_item import WorkspaceItem
from azure.quantum.job import Job

__all__ = ["Session", "SessionHost", "SessionDetails", "SessionStatus", "SessionJobFailurePolicy"]

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace
    from azure.quantum.workspace import Target


class Session(WorkspaceItem):
    """Azure Quantum Job Session: a logical grouping of jobs.

    :param workspace: Workspace instance to open the session on
    :type workspace: Workspace

    :param details: Session details model, containing the session id,
                    name, job_failure_policy, provider_id and target.
                    Either this parameter should be passed containing all
                    the session detail values, or the same values should be
                    passed as individual parameters.
    :type details: Optional[SessionDetails]

    :param target: The name of the target (or Target object) to open the session on.
    :type target: Union[str, Target, None]

    :param provider_id: The id of the provider to open the session on.
                        If not passed, it will be extracted from the target name.
    :type provider_id: Optional[str]

    :param id: The id of the session. If not passed, one random uuid will used.
    :type id: Optional[str]

    :param name: The name of the session.
                 If not passed, the name will be `session-{session-id}`.
    :type name: Optional[str]

    :param job_failure_policy: The policy that determines when a session would fail,
                               close and not accept further jobs.
    :type job_failure_policy: Union[str, SessionJobFailurePolicy, None]

    :raises ValueError: if details is passed along individual parameters,
                        or if required parameters are missing.
    """

    def __init__(
            self,
            workspace: "Workspace",
            details: Optional[SessionDetails] = None,
            target: Union[str, "Target", None] = None,
            provider_id: Optional[str] = None,
            id: Optional[str] = None,
            name: Optional[str] = None,
            job_failure_policy: Union[str, SessionJobFailurePolicy, None] = None,
            **kwargs):
        from azure.quantum.target import Target
        target_name = target.name if isinstance(target, Target) else target
        self._target = target if isinstance(target, Target) else None

        if ((details is not None)
            and ((isinstance(target, str)) or
                 (provider_id is not None) or
                 (id is not None) or
                 (name is not None) or
                 (job_failure_policy is not None))):
            raise ValueError("""If `details` is passed, you should not pass `target`,
                                `provider_id`, `id`, `name` or `job_failure_policy`.""")

        if (details is None) and (target is None):
            raise ValueError("If `session_details` is not passed, you should at least pass the `target`.")

        if details is None:
            import uuid
            import re
            id = id if id is not None else str(uuid.uuid1())
            name = name if name is not None else f"session-{id}"
            if provider_id is None:
                match = re.match(r"(\w+)\.", target_name)
                if match is not None:
                    provider_id = match.group(1)
            details = SessionDetails(id=id,
                                     name=name,
                                     provider_id=provider_id,
                                     target=target_name,
                                     job_failure_policy=job_failure_policy,
                                     **kwargs)

        super().__init__(
            workspace=workspace,
            details=details,
            **kwargs
        )

    @property
    def details(self) -> SessionDetails:
        """Get the session details.

        :return: The details about the session.
        :rtype: SessionDetails
        """
        return self._details

    @details.setter
    def details(self, value: SessionDetails):
        """Set session details.
        
        :param value: The details about the session
        :type value: SessionDetails
        """
        self._details = value

    @property
    def target(self) -> "Target":
        """Get the target associated with the session.

        :return: The target associated with the session.
        :rtype: Target
        """
        return self._target

    def open(self) -> "Session":
        """Opens a session, effectively creating a new session in the
           Azure Quantum service, and allowing it to accept jobs under it.

        :return: The session object with updated details after its opening.
        :rtype: Session
        """
        self.workspace.open_session(self)
        return self

    def close(self) -> "Session":
        """Closes a session, not allowing further jobs to be submitted under
           the session.

        :return: The session object with updated details after its closing.
        :rtype: Session
        """
        self.workspace.close_session(self)
        return self

    def refresh(self) -> "Session":
        """Fetches the latest session details from the Azure Quantum service.

        :return: The session object with updated details.
        :rtype: Session
        """
        self.workspace.refresh_session(self)
        return self

    def list_jobs(self) -> List[Job]:
        """Lists all jobs associated with this session.

        :return: A list of all jobs associated with this session.
        :rtype: typing.List[Job]
        """
        return self.workspace.list_session_jobs(session_id=self.id)

    def is_in_terminal_state(self) -> bool:
        """Returns True if the session is in one of the possible
           terminal states(Succeeded, Failed and Timed_Out).

        :return: True if the session is in one of the terminal states.
        :rtype: bool
        """
        return (self.details.status == SessionStatus.SUCCEEDED
                or self.details.status == SessionStatus.FAILED
                or self.details.status == SessionStatus.TIMED_OUT)

    def __enter__(self):
        """PEP 343 context manager implementation to use a session in
           a `with` block.
           This `__enter__` method is a no-op.
        """
        return self

    def __exit__(self, type, value, traceback):
        """PEP 343 context manager implementation to use a session in
           a `with` block.
           This `__exit__` attempts to close the session.

        :raises Exception: re-raises the exception that was caught
                           in the `with` block.
        """
        self.close()
        if isinstance(value, Exception):
            raise


class SessionHost(Protocol):
    """A protocol to allow other objects to "host" a session.
    For example, a target object can host an open session and
    have all jobs that are being submitted through it to be associated
    with that session.

    Example (job 1 to 3 will be associated the session "MySession"):

    .. highlight:: python
    .. code-block::

       with target.open_session(name="MySession") as session:
            job1 = target.submit(input_data=input_data, job_name="Job 1")
            job2 = target.submit(input_data=input_data, job_name="Job 2")
            job3 = target.submit(input_data=input_data, job_name="Job 3")
    
    """

    _latest_session: Optional[Session] = None

    @property
    def latest_session(self) -> Optional[Session]:
        """Get the latest (open) session associated with this object.

        :return: The latest session object.
        :rtype: typing.Optional[Session]
        """
        return self._latest_session

    @latest_session.setter
    def latest_session(self, session: Optional[Session]):
        """Set the latest session.
        :param value: The latest session
        :type value: Optional[Session]
        """
        self._latest_session = session

    def get_latest_session_id(self) -> Optional[str]:
        """Get the latest (open) session id associated with this object.
           This id is used to associate jobs to the latest (open) session.

        :return: The latest session id.
        :rtype: typing.Optional[str]
        """
        return self.latest_session.id if self.latest_session else None

    @abstractmethod
    def _get_azure_workspace(self) -> "Workspace":
        raise NotImplementedError

    @abstractmethod
    def _get_azure_target_id(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _get_azure_provider_id(self) -> str:
        raise NotImplementedError

    def open_session(
        self,
        details: Optional[SessionDetails] = None,
        id: Optional[str] = None,
        name: Optional[str] = None,
        job_failure_policy: Union[str, SessionJobFailurePolicy, None] = None,
        **kwargs
    ) -> Session:
        """Opens a session and associates all future job submissions to that
           session until the session is closed (which happens automatically
           after exiting a `with` block).

        Example (job 1 to 3 will be associated the session "MySession"):

        .. highlight:: python
        .. code-block::

           with target.open_session(name="MySession") as session:
                job1 = target.submit(input_data=input_data, job_name="Job 1")
                job2 = target.submit(input_data=input_data, job_name="Job 2")
                job3 = target.submit(input_data=input_data, job_name="Job 3")

        Note: If the session host (usually a `target` or qiskit `backend`)
        already has a session associated with it (in the `latest_session` property),
        then this method will first attempt to close that session before opening
        a new one.

        :param details: Session details model, containing the session id,
                        name, job_failure_policy, provider_id and target.
                        Either this parameter should be passed containing all
                        the session detail values, the same values should be
                        passed as individual parameters.

        :param id: The id of the session. If not passed, one random uuid will used.
        :type id: Optional[str]

        :param name: The name of the session.
                    If not passed, the name will be `session-{session-id}`.
        :type name: Optional[str]

        :param job_failure_policy: The policy that determines when a session would fail,
                                close and not accept further jobs.

        :return: The session object with updated details after its opening.
        :rtype: Session
        """
        if self.latest_session:
            self.latest_session.close()

        session = Session(details=details,
                          id=id,
                          name=name,
                          job_failure_policy=job_failure_policy,
                          workspace=self._get_azure_workspace(),
                          target=self._get_azure_target_id(),
                          provider_id=self._get_azure_provider_id(),
                          **kwargs)
        self.latest_session = session
        return session.open()



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
from typing import Any, Dict
from azure.core import exceptions
from azure.storage.blob import (
    BlobServiceClient,
    ContainerClient,
    BlobClient,
    BlobSasPermissions,
    ContentSettings,
    generate_blob_sas,
    generate_container_sas,
    BlobType,
    BlobProperties
)
from datetime import datetime, timedelta
from enum import Enum

logger = logging.getLogger(__name__)


def create_container(
    connection_string: str, container_name: str
) -> ContainerClient:
    """
    Creates and initialize a container; returns the client needed to access it.
    """
    blob_service_client = BlobServiceClient.from_connection_string(
        connection_string
    )
    logger.info(
        f'{"Initializing storage client for account:"}'
        + f"{blob_service_client.account_name}"
    )

    container_client = blob_service_client.get_container_client(container_name)
    create_container_using_client(container_client)
    return container_client


def create_container_using_client(container_client: ContainerClient):
    """
    Creates the container if it doesn't already exist.
    """
    if not container_client.exists():
        logger.debug(
            f'{"  - uploading to **new** container:"}'
            f"{container_client.container_name}"
        )
        container_client.create_container()


def get_container_uri(connection_string: str, container_name: str) -> str:
    """
    Creates and initialize a container;
    returns a URI with a SAS read/write token to access it.
    """
    container = create_container(connection_string, container_name)
    logger.info(
        f'{"Creating SAS token for container"}'
        + f"'{container_name}' on account: '{container.account_name}'"
    )

    sas_token = generate_container_sas(
        container.account_name,
        container.container_name,
        account_key=container.credential.account_key,
        permission=BlobSasPermissions(
            read=True, add=True, write=True, create=True
        ),
        expiry=datetime.utcnow() + timedelta(days=14),
    )

    uri = container.url + "?" + sas_token
    logger.debug(f"  - container url: '{uri}'.")
    return uri


def upload_blob(
    container: ContainerClient,
    blob_name: str,
    content_type: str,
    content_encoding: str,
    data: Any,
    return_sas_token: bool = True,
) -> str:
    """
    Uploads the given data to a blob record.
    If a blob with the given name already exist, it throws an error.

    Returns a uri with a SAS token to access the newly created blob.
    """
    create_container_using_client(container)
    logger.info(
        f"Uploading blob '{blob_name}'"
        + f"to container '{container.container_name}'"
        + f"on account: '{container.account_name}'"
    )

    content_settings = ContentSettings(
        content_type=content_type, content_encoding=content_encoding
    )
   
    blob = container.get_blob_client(blob_name)
    
    blob.upload_blob(data, content_settings=content_settings)
    logger.debug(f"  - blob '{blob_name}' uploaded. generating sas token.")

    if return_sas_token:
        uri = get_blob_uri_with_sas_token(blob)
    else:
        uri = remove_sas_token(blob.url)
    logger.debug(f"  - blob access url: '{uri}'.")

    return uri


def append_blob(
    container: ContainerClient,
    blob_name: str,
    content_type: str,
    content_encoding: str,
    data: Any,
    return_sas_token: bool = True,
    metadata: Dict[str, str] = None,
) -> str:
    """
    Uploads the given data to a blob record.
    If a blob with the given name already exist, it throws an error.

    Returns a uri with a SAS token to access the newly created blob.
    """
    create_container_using_client(container)
    logger.info(
        f"Appending data to blob '{blob_name}'"
        + f"in container '{container.container_name}'"
        + f"on account: '{container.account_name}'"
    )

    content_settings = ContentSettings(
        content_type=content_type, content_encoding=content_encoding
    )
    blob = container.get_blob_client(blob_name)
    try:
        props = blob.get_blob_properties()
        if props.blob_type != BlobType.AppendBlob:
            raise Exception("blob must be an append blob")
    except exceptions.ResourceNotFoundError:
        props = blob.create_append_blob(
            content_settings=content_settings, metadata=metadata
        )

    blob.append_block(data, len(data))
    logger.debug(f"  - blob '{blob_name}' appended. generating sas token.")

    if return_sas_token:
        uri = get_blob_uri_with_sas_token(blob)
    else:
        uri = remove_sas_token(blob.url)

    logger.debug(f"  - blob access url: '{uri}'.")

    return uri


def get_blob_uri_with_sas_token(blob: BlobClient):
    """Returns a URI for the given blob that contains a SAS Token"""
    sas_token = generate_blob_sas(
        blob.account_name,
        blob.container_name,
        blob.blob_name,
        account_key=blob.credential.account_key,
        permission=BlobSasPermissions(read=True),
        expiry=datetime.utcnow() + timedelta(days=14),
    )

    return blob.url + "?" + sas_token


def download_blob(blob_url: str) -> Any:
    """
    Downloads the given blob from the container.
    """
    blob_client = BlobClient.from_blob_url(blob_url)
    logger.info(
        f"Downloading blob '{blob_client.blob_name}'"
        + f"from container '{blob_client.container_name}'"
        + f"on account: '{blob_client.account_name}'"
    )

    response = blob_client.download_blob().readall()
    logger.debug(response)

    return response


def download_blob_properties(blob_url: str) -> BlobProperties:
    """Downloads the blob properties from Azure for the given blob URI"""
    blob_client = BlobClient.from_blob_url(blob_url)
    logger.info(
        f"Downloading blob properties '{blob_client.blob_name}'"
        + f"from container '{blob_client.container_name}'"
        + f"on account: '{blob_client.account_name}'"
    )

    response = blob_client.get_blob_properties()
    logger.debug(response)

    return response


def download_blob_metadata(blob_url: str) -> Dict[str, str]:
    """Downloads the blob metadata from the
    blob properties in Azure for the given blob URI"""
    return download_blob_properties(blob_url).metadata


def set_blob_metadata(blob_url: str, metadata: Dict[str, str]):
    """Sets the provided dictionary as the metadata on the Azure blob"""
    blob_client = BlobClient.from_blob_url(blob_url)
    logger.info(
        f"Setting blob properties '{blob_client.blob_name}'"
        + f"from container '{blob_client.container_name}' on account:"
        + f"'{blob_client.account_name}'"
    )
    return blob_client.set_blob_metadata(metadata=metadata)


def remove_sas_token(sas_uri: str) -> str:
    """Removes the SAS Token from the given URI if it contains one"""
    index = sas_uri.find("?")
    if index != -1:
        sas_uri = sas_uri[0:index]

    return sas_uri


def init_blob_for_streaming_upload(
    container: ContainerClient,
    blob_name: str,
    content_type: str,
    content_encoding: str,
    data: Any,
    return_sas_token: bool = True,
) -> str:
    """
    Uploads the given data to a blob record.
    If a blob with the given name already exist, it throws an error.

    Returns a uri with a SAS token to access the newly created blob.
    """
    create_container_using_client(container)
    logger.info(
        f"Streaming blob '{blob_name}'"
        + f"to container '{container.container_name}' on account:"
        + f"'{container.account_name}'"
    )

    content_settings = ContentSettings(
        content_type=content_type, content_encoding=content_encoding
    )
    blob = container.get_blob_client(blob_name)
    blob.stage_block()
    blob.commit_block_list()
    blob.upload_blob(data, content_settings=content_settings)
    logger.debug(f"  - blob '{blob_name}' uploaded. generating sas token.")

    if return_sas_token:
        sas_token = generate_blob_sas(
            blob.account_name,
            blob.container_name,
            blob.blob_name,
            account_key=blob.credential.account_key,
            permission=BlobSasPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(days=14),
        )

        uri = blob.url + "?" + sas_token
    else:
        uri = remove_sas_token(blob.url)

    logger.debug(f"  - blob access url: '{uri}'.")

    return uri


class StreamedBlobState(str, Enum):
    not_initialized = 0
    uploading = 1
    committed = 2


class StreamedBlob:
    """Class that provides a state machine for writing
    blobs using the Azure Block Blob API

    Internally implements a state machine for uploading blob data.
    To use, start calling `upload_data()`
    to add data blocks. Each call to `upload_data()`
    will synchronously upload an individual block to Azure.
    Once all blocks have been added, call `commit()`
    to commit the blocks and make the blob available/readable.

    :param container: The container client that the blob will be uploaded to
    :param blob_name: The name of the blob
        (including optional path) within the blob container
    :param content_type: The HTTP content type to apply to the blob metadata
    :param content_encoding: The HTTP
        content encoding to apply to the blob metadata
    """

    def __init__(
        self,
        container: ContainerClient,
        blob_name: str,
        content_type: str,
        content_encoding: str,
    ):
        self.container = container
        self.blob_name = blob_name
        self.content_settings = ContentSettings(
            content_type=content_type, content_encoding=content_encoding
        )
        self.state = StreamedBlobState.not_initialized
        self.blob = container.get_blob_client(blob_name)
        self.blocks = []

    def upload_data(self, data):
        """Synchronously uploads a block to the given block blob in Azure

        :param data: The data to be uploaded as a block.
        :type data: Union[Iterable[AnyStr], IO[AnyStr]]
        """
        if self.state == StreamedBlobState.not_initialized:
            create_container_using_client(self.container)
            logger.info(
                f"Streaming blob '{self.blob_name}' to container"
                + f"'{self.container.container_name}'"
                + f"on account: '{self.container.account_name}'"
            )
            self.initialized = True

        self.state = StreamedBlobState.uploading
        id = self._get_next_block_id()
        logger.debug(f"Uploading block '{id}' to {self.blob_name}")
        self.blob.stage_block(id, data, length=len(data))
        self.blocks.append(id)

    def commit(self, metadata: Dict[str, str] = None):
        """Synchronously commits all previously
        uploaded blobs to the block blob

        :param metadata: Optional dictionary of
               metadata to be applied to the block blob
        """
        if self.state == StreamedBlobState.not_initialized:
            raise Exception("StreamedBlob cannot commit before uploading data")
        elif self.state == StreamedBlobState.committed:
            raise Exception("StreamedBlob is already committed")

        logger.debug(f"Committing {len(self.blocks)} blocks {self.blob_name}")
        self.blob.commit_block_list(
            self.blocks,
            content_settings=self.content_settings,
            metadata=metadata,
        )
        self.state = StreamedBlobState.committed
        logger.debug(f"Committed {self.blob_name}")

    def getUri(self, with_sas_token: bool = False):
        """Gets the full Azure Storage URI for the
        uploaded blob after it has been committed"""
        if self.state != StreamedBlobState.committed:
            raise Exception("Can only retrieve sas token for committed blob")
        if with_sas_token:
            return get_blob_uri_with_sas_token(self.blob)

        return remove_sas_token(self.blob.url)

    def _get_next_block_id(self):
        return f"{len(self.blocks):10}"



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

"""Defines classes for interacting with Microsoft Estimator"""

__all__ = ["ErrorBudgetPartition", "MicrosoftEstimator",
           "MicrosoftEstimatorJob", "MicrosoftEstimatorResult",
           "MicrosoftEstimatorParams", "QECScheme", "QubitParams"]

from .job import MicrosoftEstimatorJob
from .result import MicrosoftEstimatorResult
from .target import ErrorBudgetPartition, MicrosoftEstimator, \
    MicrosoftEstimatorParams, QECScheme, QubitParams



##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

"""Defines set of targets for interacting with Azure Quantum"""

from .target import Target
from .ionq import IonQ
from .quantinuum import Quantinuum
from .rigetti import Rigetti
from .pasqal import Pasqal
from .microsoft.elements.dft import MicrosoftElementsDft, MicrosoftElementsDftJob

# Default targets to use when there is no target class
# associated with a given target ID
DEFAULT_TARGETS = {
    "ionq": IonQ,
    "quantinuum": Quantinuum,
    "rigetti": Rigetti,
    "pasqal": Pasqal
}


__all__ = [
    "Target",
    "IonQ",
    "Quantinuum",
    "DEFAULT_TARGETS"
    ]


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from ... import Job
from ...job.base_job import DEFAULT_TIMEOUT
from ..._client.models import JobDetails
from .result import MicrosoftEstimatorResult

class MicrosoftEstimatorJob(Job):
    """
    A dedicated job class for jobs from the microsoft.estimator target.
    """

    def __init__(self, workspace, job_details: JobDetails, **kwargs):
        super().__init__(workspace, job_details, **kwargs)

    def get_results(self, timeout_secs: float = DEFAULT_TIMEOUT) -> MicrosoftEstimatorResult:
        """
        Get estimation result.

        :param timeout_secs: Timeout in seconds, defaults to 300 sec.
        :type timeout_secs: float
        :return: Estimation result
        :rtype: MicrosoftEstimatorResult
        """
        try:
            results = super().get_results(timeout_secs)
            return MicrosoftEstimatorResult(results)
        except RuntimeError:
            error_obj = self.details.error_data
            message = "Cannot retrieve results as job execution failed " \
                      f"({error_obj.code}: {error_obj.message})"
            raise RuntimeError(message)

